/*
 * Copyright 2019 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package androidx.ui.text

import android.graphics.Bitmap
import androidx.test.filters.SmallTest
import androidx.test.platform.app.InstrumentationRegistry
import androidx.ui.core.Constraints
import androidx.ui.core.Density
import androidx.ui.core.PxPosition
import androidx.ui.core.ipx
import androidx.ui.core.px
import androidx.ui.core.sp
import androidx.ui.core.withDensity
import androidx.ui.engine.geometry.Rect
import androidx.ui.engine.geometry.Size
import androidx.ui.text.FontTestData.Companion.BASIC_MEASURE_FONT
import androidx.ui.text.style.TextDirection
import androidx.ui.text.font.asFontFamily
import androidx.ui.graphics.Color
import androidx.ui.text.matchers.equalToBitmap
import androidx.ui.painting.Canvas
import androidx.ui.painting.Paint
import androidx.ui.text.style.TextOverflow
import com.google.common.truth.Truth.assertThat
import org.junit.Assert
import org.junit.Test
import org.junit.runner.RunWith
import org.junit.runners.JUnit4
import kotlin.math.ceil

@RunWith(JUnit4::class)
@SmallTest
class TextPainterIntegrationTest {

    private val fontFamily = BASIC_MEASURE_FONT.asFontFamily()
    private val density = Density(density = 1f)
    private val context = InstrumentationRegistry.getInstrumentation().context
    private val resourceLoader = TestFontResourceLoader(context)

    @Test
    fun preferredLineHeight_style_set() {
        withDensity(density) {
            val fontSize = 20.sp
            val textStyle = TextStyle(fontSize = fontSize, fontFamily = fontFamily)
            val textPainter = TextPainter(
                style = textStyle,
                density = density,
                resourceLoader = resourceLoader
            )
            val preferredHeight = textPainter.preferredLineHeight

            assertThat(preferredHeight).isEqualTo(fontSize.toPx().value)
        }
    }

    // TODO(Migration/qqd): The default font size should be 14.0 but it returns 15.0. Need further
    // investigation. It is being changed in the native level, and probably related to the font.
//    @Test
//    fun preferredLineHeight_style_not_set() {
//        val defaultTextStyle = TextStyle(fontFamily = fontFamily)
//        val textPainter = TextPainter(style = defaultTextStyle)
//
//        val prefferedHeight = textPainter.preferredLineHeight
//
//        assertThat(prefferedHeight).isEqualTo(14.0)
//    }

    @Test
    fun minIntrinsicWidth_getter() {
        val fontSize = 20.sp
        val text = "Hello"
        val textStyle = TextStyle(fontSize = fontSize, fontFamily = fontFamily)
        val annotatedString = AnnotatedString(
            text = text,
            textStyles = listOf(AnnotatedString.Item(textStyle, 0, text.length))
        )
        val textPainter = TextPainter(
            text = annotatedString,
            paragraphStyle = ParagraphStyle(textDirection = TextDirection.Rtl),
            density = density,
            resourceLoader = resourceLoader
        )

        textPainter.layout(Constraints())

        assertThat(textPainter.minIntrinsicWidth).isEqualTo(0.0f)
    }

    @Test
    fun maxIntrinsicWidth_getter() {
        withDensity(density) {
            val fontSize = 20.sp
            val text = "Hello"
            val textStyle = TextStyle(fontSize = fontSize, fontFamily = fontFamily)
            val annotatedString = AnnotatedString(
                text = text,
                textStyles = listOf(
                    AnnotatedString.Item(
                        textStyle,
                        0,
                        text.length
                    )
                )
            )
            val textPainter = TextPainter(
                text = annotatedString,
                paragraphStyle = ParagraphStyle(textDirection = TextDirection.Rtl),
                density = density,
                resourceLoader = resourceLoader
            )

            textPainter.layout(Constraints())

            assertThat(textPainter.maxIntrinsicWidth).isEqualTo(fontSize.toPx().value * text.length)
        }
    }

    @Test
    fun width_getter() {
        withDensity(density) {
            val fontSize = 20.sp
            val text = "Hello"
            val textStyle = TextStyle(fontSize = fontSize, fontFamily = fontFamily)
            val annotatedString = AnnotatedString(
                text = text,
                textStyles = listOf(
                    AnnotatedString.Item(
                        textStyle,
                        0,
                        text.length
                    )
                )
            )
            val textPainter = TextPainter(
                text = annotatedString,
                paragraphStyle = ParagraphStyle(textDirection = TextDirection.Rtl),
                density = density,
                resourceLoader = resourceLoader
            )

            textPainter.layout(Constraints(0.ipx, 200.ipx))

            assertThat(textPainter.width).isEqualTo(fontSize.toPx().value * text.length)
        }
    }

    @Test
    fun width_getter_with_small_width() {
        val text = "Hello"
        val width = 80.ipx
        val textStyle = TextStyle(fontSize = 20.sp, fontFamily = fontFamily)
        val annotatedString = AnnotatedString(
            text = text,
            textStyles = listOf(AnnotatedString.Item(textStyle, 0, text.length))
        )
        val textPainter = TextPainter(
            text = annotatedString,
            paragraphStyle = ParagraphStyle(textDirection = TextDirection.Rtl),
            density = density,
            resourceLoader = resourceLoader
        )

        textPainter.layout(Constraints(maxWidth = width))

        assertThat(textPainter.width).isEqualTo(width.value.toFloat())
    }

    @Test
    fun height_getter() {
        withDensity(density) {
            val fontSize = 20.sp
            val textStyle = TextStyle(fontSize = fontSize, fontFamily = fontFamily)
            val text = "hello"
            val annotatedString = AnnotatedString(
                text = text,
                textStyles = listOf(
                    AnnotatedString.Item(
                        textStyle,
                        0,
                        text.length
                    )
                )
            )
            val textPainter = TextPainter(
                text = annotatedString,
                paragraphStyle = ParagraphStyle(textDirection = TextDirection.Rtl),
                density = density,
                resourceLoader = resourceLoader
            )

            textPainter.layout(Constraints())

            assertThat(textPainter.height).isEqualTo(fontSize.toPx().value)
        }
    }

    @Test
    fun size_getter() {
        withDensity(density) {
            val fontSize = 20.sp
            val fontSizeInPx = fontSize.toPx().value
            val text = "Hello"
            val textStyle = TextStyle(fontSize = fontSize, fontFamily = fontFamily)
            val annotatedString = AnnotatedString(
                text = text,
                textStyles = listOf(
                    AnnotatedString.Item(
                        textStyle,
                        0,
                        text.length
                    )
                )
            )
            val textPainter = TextPainter(
                text = annotatedString,
                paragraphStyle = ParagraphStyle(textDirection = TextDirection.Rtl),
                density = density,
                resourceLoader = resourceLoader
            )

            textPainter.layout(Constraints())

            assertThat(textPainter.size).isEqualTo(
                Size(
                    width = fontSizeInPx * text.length,
                    height = fontSizeInPx
                )
            )
        }
    }

    @Test
    fun didExceedMaxLines_exceed() {
        var text = ""
        for (i in 1..50) text += " Hello"
        val annotatedString = AnnotatedString(text = text)
        val textPainter = TextPainter(
            text = annotatedString,
            paragraphStyle = ParagraphStyle(textDirection = TextDirection.Rtl),
            maxLines = 2,
            density = density,
            resourceLoader = resourceLoader
        )

        textPainter.layout(Constraints(0.ipx, 200.ipx))

        assertThat(textPainter.didExceedMaxLines).isTrue()
    }

    @Test
    fun didExceedMaxLines_not_exceed() {
        val text = "Hello"
        val annotatedString = AnnotatedString(text = text)
        val textPainter = TextPainter(
            text = annotatedString,
            paragraphStyle = ParagraphStyle(textDirection = TextDirection.Rtl),
            maxLines = 2,
            density = density,
            resourceLoader = resourceLoader
        )

        textPainter.layout(Constraints(0.ipx, 200.ipx))

        assertThat(textPainter.didExceedMaxLines).isFalse()
    }

    @Test
    fun layout_build_paragraph() {
        val textPainter = TextPainter(
            text = AnnotatedString(text = "Hello"),
            paragraphStyle = ParagraphStyle(textDirection = TextDirection.Ltr),
            density = density,
            resourceLoader = resourceLoader
        )

        textPainter.layout(Constraints(0.ipx, 20.ipx))

        assertThat(textPainter.multiParagraph).isNotNull()
    }

    @Test
    fun getPositionForOffset_First_Character() {
        val text = "Hello"
        val annotatedString = AnnotatedString(
            text = text,
            textStyles = listOf(
                AnnotatedString.Item(
                    TextStyle(fontSize = 20.sp, fontFamily = fontFamily),
                    0,
                    text.length
                )
            )
        )
        val textPainter = TextPainter(
            text = annotatedString,
            paragraphStyle = ParagraphStyle(textDirection = TextDirection.Ltr),
            density = density,
            resourceLoader = resourceLoader
        )
        textPainter.layout(Constraints())

        val selection = textPainter.getOffsetForPosition(PxPosition.Origin)

        assertThat(selection).isEqualTo(0)
    }

    @Test
    fun getPositionForOffset_other_Character() {
        withDensity(density) {
            val fontSize = 20.sp
            val characterIndex = 2 // Start from 0.
            val text = "Hello"
            val annotatedString = AnnotatedString(
                text = text,
                textStyles = listOf(
                    AnnotatedString.Item(
                        TextStyle(fontSize = fontSize, fontFamily = fontFamily),
                        0,
                        text.length
                    )
                )
            )
            val textPainter = TextPainter(
                text = annotatedString,
                paragraphStyle = ParagraphStyle(textDirection = TextDirection.Ltr),
                density = density,
                resourceLoader = resourceLoader
            )
            textPainter.layout(Constraints())

            val selection = textPainter.getOffsetForPosition(
                position = PxPosition((fontSize.toPx().value * characterIndex + 1).px, 0.px)
            )

            assertThat(selection).isEqualTo(characterIndex)
        }
    }

    @Test
    fun hasOverflowShaderFalse() {
        val text = "Hello"
        val textStyle = TextStyle(fontSize = 20.sp, fontFamily = fontFamily)
        val annotatedString = AnnotatedString(
            text = text,
            textStyles = listOf(AnnotatedString.Item(textStyle, 0, text.length))
        )
        val textPainter = TextPainter(
            text = annotatedString,
            paragraphStyle = ParagraphStyle(textDirection = TextDirection.Ltr),
            density = density,
            resourceLoader = resourceLoader
        )

        textPainter.layout(Constraints())

        assertThat(textPainter.hasVisualOverflow).isFalse()
    }

    @Test
    fun hasOverflowShaderFadeHorizontallyTrue() {
        var text = ""
        for (i in 1..15) {
            text = text + "Hello World"
        }
        val textStyle = TextStyle(fontSize = 20.sp, fontFamily = fontFamily)
        val annotatedString = AnnotatedString(
            text = text,
            textStyles = listOf(AnnotatedString.Item(textStyle, 0, text.length))
        )
        val textPainter = TextPainter(
            text = annotatedString,
            paragraphStyle = ParagraphStyle(textDirection = TextDirection.Ltr),
            overflow = TextOverflow.Fade,
            softWrap = false,
            maxLines = 1,
            density = density,
            resourceLoader = resourceLoader
        )

        textPainter.layout(Constraints(maxWidth = 100.ipx))

        assertThat(textPainter.hasVisualOverflow).isTrue()
    }

    @Test
    fun hasOverflowShaderFadeVerticallyTrue() {
        var text = ""
        for (i in 1..30) {
            text = text + "Hello World"
        }
        val textStyle = TextStyle(fontSize = 20.sp, fontFamily = fontFamily)
        val annotatedString = AnnotatedString(
            text = text,
            textStyles = listOf(AnnotatedString.Item(textStyle, 0, text.length))
        )
        val textPainter = TextPainter(
            text = annotatedString,
            paragraphStyle = ParagraphStyle(textDirection = TextDirection.Ltr),
            overflow = TextOverflow.Fade,
            maxLines = 2,
            density = density,
            resourceLoader = resourceLoader
        )

        textPainter.layout(Constraints(maxWidth = 100.ipx))

        assertThat(textPainter.hasVisualOverflow).isTrue()
    }

    @Test
    fun testBackgroundPaint_paint_wrap_multiLines() {
        withDensity(density) {
            // Setup test.
            val fontSize = 20.sp
            val fontSizeInPx = fontSize.toPx().value
            val text = "HelloHello"
            val textStyle = TextStyle(fontSize = fontSize, fontFamily = fontFamily)
            val annotatedString = AnnotatedString(
                text = text,
                textStyles = listOf(
                    AnnotatedString.Item(
                        textStyle,
                        0,
                        text.length
                    )
                )
            )
            val textPainter = TextPainter(
                text = annotatedString,
                paragraphStyle = ParagraphStyle(textDirection = TextDirection.Ltr),
                density = density,
                resourceLoader = resourceLoader
            )
            textPainter.layout(Constraints(maxWidth = 120.ipx))

            val expectedBitmap = Bitmap.createBitmap(
                ceil(textPainter.width).toInt(),
                ceil(textPainter.height).toInt(),
                Bitmap.Config.ARGB_8888
            )
            val expectedCanvas =
                Canvas(android.graphics.Canvas(expectedBitmap))
            val expectedPaint = Paint()
            val defaultSelectionColor = Color(0x6633B5E5)
            expectedPaint.color = defaultSelectionColor

            val firstLineLeft = textPainter.multiParagraph?.getLineLeft(0)
            val secondLineLeft = textPainter.multiParagraph?.getLineLeft(1)
            val firstLineRight = textPainter.multiParagraph?.getLineRight(0)
            val secondLineRight = textPainter.multiParagraph?.getLineRight(1)
            expectedCanvas.drawRect(
                Rect(firstLineLeft!!, 0f, firstLineRight!!, fontSizeInPx),
                expectedPaint
            )
            expectedCanvas.drawRect(
                Rect(
                    secondLineLeft!!,
                    fontSizeInPx,
                    secondLineRight!!,
                    textPainter.multiParagraph!!.height
                ),
                expectedPaint
            )

            val actualBitmap = Bitmap.createBitmap(
                ceil(textPainter.width).toInt(),
                ceil(textPainter.height).toInt(),
                Bitmap.Config.ARGB_8888
            )
            val actualCanvas = Canvas(android.graphics.Canvas(actualBitmap))

            // Run.
            // Select all.
            textPainter.paintBackground(
                start = 0,
                end = text.length,
                color = defaultSelectionColor,
                canvas = actualCanvas
            )

            // Assert.
            Assert.assertThat(actualBitmap, equalToBitmap(expectedBitmap))
        }
    }

    @Test
    fun testBackgroundPaint_paint_with_default_color() {
        withDensity(density) {
            // Setup test.
            val selectionStart = 0
            val selectionEnd = 3
            val fontSize = 20.sp
            val fontSizeInPx = fontSize.toPx().value
            val text = "Hello"
            val textStyle = TextStyle(fontSize = fontSize, fontFamily = fontFamily)
            val annotatedString = AnnotatedString(
                text = text,
                textStyles = listOf(
                    AnnotatedString.Item(
                        textStyle,
                        0,
                        text.length
                    )
                )
            )
            val textPainter = TextPainter(
                text = annotatedString,
                paragraphStyle = ParagraphStyle(textDirection = TextDirection.Ltr),
                density = density,
                resourceLoader = resourceLoader
            )
            textPainter.layout(Constraints())

            val expectedBitmap = Bitmap.createBitmap(
                ceil(textPainter.width).toInt(),
                ceil(textPainter.height).toInt(),
                Bitmap.Config.ARGB_8888
            )
            val expectedCanvas =
                Canvas(android.graphics.Canvas(expectedBitmap))
            val expectedPaint = Paint()
            val defaultSelectionColor = Color(0x6633B5E5)
            expectedPaint.color = defaultSelectionColor
            expectedCanvas.drawRect(
                Rect(
                    left = 0f,
                    top = 0f,
                    right = fontSizeInPx * (selectionEnd - selectionStart),
                    bottom = fontSizeInPx
                ),
                expectedPaint
            )

            val actualBitmap = Bitmap.createBitmap(
                ceil(textPainter.width).toInt(),
                ceil(textPainter.height).toInt(),
                Bitmap.Config.ARGB_8888
            )
            val actualCanvas = Canvas(android.graphics.Canvas(actualBitmap))

            // Run.
            textPainter.paintBackground(
                start = selectionStart,
                end = selectionEnd,
                color = defaultSelectionColor,
                canvas = actualCanvas
            )

            // Assert
            Assert.assertThat(actualBitmap, equalToBitmap(expectedBitmap))
        }
    }

    @Test
    fun testBackgroundPaint_paint_with_default_color_bidi() {
        withDensity(density) {
            // Setup test.
            val textLTR = "Hello"
            // From right to left: שלום
            val textRTL = "\u05e9\u05dc\u05d5\u05dd"
            val text = textLTR + textRTL
            val selectionLTRStart = 2
            val selectionRTLEnd = 2
            val fontSize = 20.sp
            val fontSizeInPx = fontSize.toPx().value
            val textStyle = TextStyle(fontSize = fontSize, fontFamily = fontFamily)
            val annotatedString = AnnotatedString(
                text = text,
                textStyles = listOf(
                    AnnotatedString.Item(
                        textStyle,
                        0,
                        text.length
                    )
                )
            )
            val textPainter = TextPainter(
                text = annotatedString,
                paragraphStyle = ParagraphStyle(textDirection = TextDirection.Ltr),
                density = density,
                resourceLoader = resourceLoader
            )
            textPainter.layout(Constraints())

            val expectedBitmap = Bitmap.createBitmap(
                ceil(textPainter.width).toInt(),
                ceil(textPainter.height).toInt(),
                Bitmap.Config.ARGB_8888
            )
            val expectedCanvas =
                Canvas(android.graphics.Canvas(expectedBitmap))
            val expectedPaint = Paint()
            val defaultSelectionColor = Color(0x6633B5E5)
            expectedPaint.color = defaultSelectionColor
            // Select "llo".
            expectedCanvas.drawRect(
                Rect(
                    left = fontSizeInPx * selectionLTRStart,
                    top = 0f,
                    right = textLTR.length * fontSizeInPx,
                    bottom = fontSizeInPx
                ),
                expectedPaint
            )

            // Select "של"
            expectedCanvas.drawRect(
                Rect(
                    left = (textLTR.length + textRTL.length - selectionRTLEnd) * fontSizeInPx,
                    top = 0f,
                    right = (textLTR.length + textRTL.length) * fontSizeInPx,
                    bottom = fontSizeInPx
                ),
                expectedPaint
            )

            val actualBitmap = Bitmap.createBitmap(
                ceil(textPainter.width).toInt(),
                ceil(textPainter.height).toInt(),
                Bitmap.Config.ARGB_8888
            )
            val actualCanvas = Canvas(android.graphics.Canvas(actualBitmap))

            // Run.
            textPainter.paintBackground(
                start = selectionLTRStart,
                end = textLTR.length + selectionRTLEnd,
                color = defaultSelectionColor,
                canvas = actualCanvas
            )

            // Assert
            Assert.assertThat(actualBitmap, equalToBitmap(expectedBitmap))
        }
    }

    @Test
    fun testBackgroundPaint_paint_with_customized_color() {
        withDensity(density) {
            // Setup test.
            val selectionStart = 0
            val selectionEnd = 3
            val fontSize = 20.sp
            val fontSizeInPx = fontSize.toPx().value
            val text = "Hello"
            val textStyle = TextStyle(fontSize = fontSize, fontFamily = fontFamily)
            val annotatedString = AnnotatedString(
                text = text,
                textStyles = listOf(
                    AnnotatedString.Item(
                        textStyle,
                        0,
                        text.length
                    )
                )
            )
            val selectionColor = Color(0x66AABB33)
            val textPainter = TextPainter(
                text = annotatedString,
                paragraphStyle = ParagraphStyle(textDirection = TextDirection.Ltr),
                density = density,
                resourceLoader = resourceLoader
            )
            textPainter.layout(Constraints())

            val expectedBitmap = Bitmap.createBitmap(
                ceil(textPainter.width).toInt(),
                ceil(textPainter.height).toInt(),
                Bitmap.Config.ARGB_8888
            )
            val expectedCanvas =
                Canvas(android.graphics.Canvas(expectedBitmap))
            val expectedPaint = Paint()
            expectedPaint.color = selectionColor
            expectedCanvas.drawRect(
                Rect(
                    left = 0f,
                    top = 0f,
                    right = fontSizeInPx * (selectionEnd - selectionStart),
                    bottom = fontSizeInPx
                ),
                expectedPaint
            )

            val actualBitmap = Bitmap.createBitmap(
                ceil(textPainter.width).toInt(),
                ceil(textPainter.height).toInt(),
                Bitmap.Config.ARGB_8888
            )
            val actualCanvas = Canvas(android.graphics.Canvas(actualBitmap))

            // Run.
            textPainter.paintBackground(
                start = selectionStart,
                end = selectionEnd,
                color = selectionColor,
                canvas = actualCanvas
            )

            // Assert
            Assert.assertThat(actualBitmap, equalToBitmap(expectedBitmap))
        }
    }
}